#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Dict, List

###############################################################################

# The order of the readers in this impl dict is important.
#
# Example:
# if TiffReader was placed before OmeTiffReader,
# we would never hit the OmeTiffReader

# Additionally while so many formats can be read by base-imageio
# Our custom reader may be more well-suited for interactions
# Example:
# DefaultReader supports LSM, and all the similar "Tiff-Like"
# formats, but TiffReader does as well and it has better
# dask chunking + metadata parsing than DefaultReader for those formats

# BASE-IMAGEIO FORMATS (with tifffile + non-existant removals)
#
# Pulled using:
# from imageio import formats
# routes = {}
# for f in formats:
#   for ext in f.extensions:
#       routes[ext[1:]] = ["aicsimageio.readers.default_reader.DefaultReader"]

FORMAT_IMPLEMENTATIONS: Dict[str, List[str]] = {
    "1sc": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "2fl": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "3fr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "acff": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "acqp": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "afi": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "afm": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "aim": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "al3d": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ali": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "am": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "amiramesh": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ano": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "apl": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "arf": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "array-like": ["aicsimageio.readers.array_like_reader.ArrayLikeReader"],
    "arw": ["aicsimageio.readers.default_reader.DefaultReader"],
    "avi": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "bay": ["aicsimageio.readers.default_reader.DefaultReader"],
    "bif": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "bin": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "bip": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "bmp": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "bmq": ["aicsimageio.readers.default_reader.DefaultReader"],
    "bsdf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "bufr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "bw": ["aicsimageio.readers.default_reader.DefaultReader"],
    "c01": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "cap": ["aicsimageio.readers.default_reader.DefaultReader"],
    "cat": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "cfg": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ch5": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "cif": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "cine": ["aicsimageio.readers.default_reader.DefaultReader"],
    "cr2": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "crw": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "cs1": ["aicsimageio.readers.default_reader.DefaultReader"],
    "csv": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ct": ["aicsimageio.readers.default_reader.DefaultReader"],
    "ct.img": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "cur": ["aicsimageio.readers.default_reader.DefaultReader"],
    "cut": ["aicsimageio.readers.default_reader.DefaultReader"],
    "cxd": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "czi": [
        "aicsimageio.readers.czi_reader.CziReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "dat": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "db": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "dc2": ["aicsimageio.readers.default_reader.DefaultReader"],
    "dcm": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "dcr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "dcx": ["aicsimageio.readers.default_reader.DefaultReader"],
    "dds": ["aicsimageio.readers.default_reader.DefaultReader"],
    "df3": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "dicom": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "dm2": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "dm3": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "dng": ["aicsimageio.readers.default_reader.DefaultReader"],
    "drf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "dsc": ["aicsimageio.readers.default_reader.DefaultReader"],
    "dti": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "dv": [
        "aicsimageio.readers.dv_reader.DVReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "ecw": ["aicsimageio.readers.default_reader.DefaultReader"],
    "emf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "eps": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "epsi": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "erf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "exp": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "exr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "fake": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "fdf": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "fff": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "ffr": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "fid": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "fit": ["aicsimageio.readers.default_reader.DefaultReader"],
    "fits": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "flc": ["aicsimageio.readers.default_reader.DefaultReader"],
    "flex": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "fli": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "fpx": ["aicsimageio.readers.default_reader.DefaultReader"],
    "frm": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ftc": ["aicsimageio.readers.default_reader.DefaultReader"],
    "fts": ["aicsimageio.readers.default_reader.DefaultReader"],
    "ftu": ["aicsimageio.readers.default_reader.DefaultReader"],
    "fz": ["aicsimageio.readers.default_reader.DefaultReader"],
    "g3": ["aicsimageio.readers.default_reader.DefaultReader"],
    "gbr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "gdcm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "gel": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "gif": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "gipl": ["aicsimageio.readers.default_reader.DefaultReader"],
    "grey": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "grib": ["aicsimageio.readers.default_reader.DefaultReader"],
    "h5": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "hdf": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "hdf5": ["aicsimageio.readers.default_reader.DefaultReader"],
    "hdp": ["aicsimageio.readers.default_reader.DefaultReader"],
    "hdr": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "hed": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "his": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "htd": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "htm": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "html": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "hx": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "i2i": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ia": ["aicsimageio.readers.default_reader.DefaultReader"],
    "icns": ["aicsimageio.readers.default_reader.DefaultReader"],
    "ico": ["aicsimageio.readers.default_reader.DefaultReader"],
    "ics": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ids": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "iff": ["aicsimageio.readers.default_reader.DefaultReader"],
    "iim": ["aicsimageio.readers.default_reader.DefaultReader"],
    "iiq": ["aicsimageio.readers.default_reader.DefaultReader"],
    "im": ["aicsimageio.readers.default_reader.DefaultReader"],
    "im3": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "img": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "imggz": ["aicsimageio.readers.default_reader.DefaultReader"],
    "ims": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "inf": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "inr": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ipl": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "ipm": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ipw": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "j2c": ["aicsimageio.readers.default_reader.DefaultReader"],
    "j2k": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "jfif": ["aicsimageio.readers.default_reader.DefaultReader"],
    "jif": ["aicsimageio.readers.default_reader.DefaultReader"],
    "jng": ["aicsimageio.readers.default_reader.DefaultReader"],
    "jp2": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "jpc": ["aicsimageio.readers.default_reader.DefaultReader"],
    "jpe": ["aicsimageio.readers.default_reader.DefaultReader"],
    "jpeg": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "jpf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "jpg": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "jpk": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "jpx": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "jxr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "k25": ["aicsimageio.readers.default_reader.DefaultReader"],
    "kc2": ["aicsimageio.readers.default_reader.DefaultReader"],
    "kdc": ["aicsimageio.readers.default_reader.DefaultReader"],
    "klb": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "koa": ["aicsimageio.readers.default_reader.DefaultReader"],
    "l2d": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "labels": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "lbm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "lei": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "lfp": ["aicsimageio.readers.default_reader.DefaultReader"],
    "lfr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "lif": [
        "aicsimageio.readers.lif_reader.LifReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "liff": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "lim": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "lms": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "lsm": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "mdb": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "mdc": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mef": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mgh": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mha": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mhd": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mic": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mkv": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mnc": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "mnc2": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mng": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "mod": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "mos": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mov": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "mp4": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mpeg": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mpg": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mpo": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mrc": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "mri": ["aicsimageio.readers.default_reader.DefaultReader"],
    "mrw": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "msp": ["aicsimageio.readers.default_reader.DefaultReader"],
    "msr": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "mtb": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "mvd2": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "naf": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "nd": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "nd2": [
        "aicsimageio.readers.nd2_reader.ND2Reader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "ndpi": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ndpis": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "nef": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "nhdr": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "nia": ["aicsimageio.readers.default_reader.DefaultReader"],
    "nii": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "nii.gz": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "niigz": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "npz": ["aicsimageio.readers.default_reader.DefaultReader"],
    "nrrd": [
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "nrw": ["aicsimageio.readers.default_reader.DefaultReader"],
    "obf": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "oib": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "oif": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "oir": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ome": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ome.tif": [
        "aicsimageio.readers.bfio_reader.OmeTiledTiffReader",
        "aicsimageio.readers.ome_tiff_reader.OmeTiffReader",
        "aicsimageio.readers.tiff_reader.TiffReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "ome.tiff": [
        "aicsimageio.readers.bfio_reader.OmeTiledTiffReader",
        "aicsimageio.readers.ome_tiff_reader.OmeTiffReader",
        "aicsimageio.readers.tiff_reader.TiffReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "orf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "par": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "pbm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "pcd": ["aicsimageio.readers.default_reader.DefaultReader"],
    "pcoraw": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "pct": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "pcx": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "pef": ["aicsimageio.readers.default_reader.DefaultReader"],
    "pfm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "pgm": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "pic": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "pict": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "png": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "pnl": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ppm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "pr3": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "ps": ["aicsimageio.readers.default_reader.DefaultReader"],
    "psd": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "ptx": ["aicsimageio.readers.default_reader.DefaultReader"],
    "pxn": ["aicsimageio.readers.default_reader.DefaultReader"],
    "pxr": ["aicsimageio.readers.default_reader.DefaultReader"],
    "qptiff": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "qtk": ["aicsimageio.readers.default_reader.DefaultReader"],
    "r3d": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "raf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "ras": ["aicsimageio.readers.default_reader.DefaultReader"],
    "raw": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "rcpnl": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "rdc": ["aicsimageio.readers.default_reader.DefaultReader"],
    "rec": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "rgb": ["aicsimageio.readers.default_reader.DefaultReader"],
    "rgba": ["aicsimageio.readers.default_reader.DefaultReader"],
    "rw2": ["aicsimageio.readers.default_reader.DefaultReader"],
    "rwl": ["aicsimageio.readers.default_reader.DefaultReader"],
    "rwz": ["aicsimageio.readers.default_reader.DefaultReader"],
    "scan": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "scn": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "sdt": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "seq": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "sif": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "sld": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "sm2": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "sm3": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "spc": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "spe": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "spi": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "sr2": ["aicsimageio.readers.default_reader.DefaultReader"],
    "srf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "srw": ["aicsimageio.readers.default_reader.DefaultReader"],
    "st": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "sti": ["aicsimageio.readers.default_reader.DefaultReader"],
    "stk": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "stp": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "svs": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "swf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "sxm": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "targa": ["aicsimageio.readers.default_reader.DefaultReader"],
    "tfr": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "tga": [
        "aicsimageio.readers.default_reader.DefaultReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
    ],
    "thm": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "tif": [
        "aicsimageio.readers.ome_tiff_reader.OmeTiffReader",
        "aicsimageio.readers.tiff_reader.TiffReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "tiff": [
        "aicsimageio.readers.ome_tiff_reader.OmeTiffReader",
        "aicsimageio.readers.tiff_reader.TiffReader",
        "aicsimageio.readers.bioformats_reader.BioformatsReader",
        "aicsimageio.readers.default_reader.DefaultReader",
    ],
    "tim": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "tnb": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "top": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "txt": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "v": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "vff": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "vms": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "vsi": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "vtk": ["aicsimageio.readers.default_reader.DefaultReader"],
    "vws": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "wap": ["aicsimageio.readers.default_reader.DefaultReader"],
    "wat": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "wav": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "wbm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "wbmp": ["aicsimageio.readers.default_reader.DefaultReader"],
    "wdp": ["aicsimageio.readers.default_reader.DefaultReader"],
    "webp": ["aicsimageio.readers.default_reader.DefaultReader"],
    "wlz": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "wmf": ["aicsimageio.readers.default_reader.DefaultReader"],
    "wmv": ["aicsimageio.readers.default_reader.DefaultReader"],
    "wpi": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "xbm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "xdce": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "xml": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "xpm": ["aicsimageio.readers.default_reader.DefaultReader"],
    "xqd": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "xqf": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "xv": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "xys": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "zfp": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "zfr": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "zip": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "zpo": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
    "zvi": ["aicsimageio.readers.bioformats_reader.BioformatsReader"],
}

READER_TO_INSTALL: Dict[str, str] = {
    "aicsimageio.readers.bioformats_reader.BioformatsReader": "bioformats_jar",
    "aicsimageio.readers.default_reader.DefaultReader": "aicsimageio[base-imageio]",
    "aicsimageio.readers.lif_reader.LifReader": "readlif>=0.6.4",
    "aicsimageio.readers.czi_reader.CziReader": "aicspylibczi>=3.1.1",
    "aicsimageio.readers.dv_reader.DVReader": "aicsimageio[dv]",
    "aicsimageio.readers.nd2_reader.ND2Reader": "aicsimageio[nd2]",
}



import math
from typing import Dict, List, Optional, Tuple

import zarr
from ome_zarr.io import parse_url
from ome_zarr.scale import Scaler
from ome_zarr.writer import write_image
from zarr.storage import default_compressor

from .. import exceptions, types
from ..metadata import utils
from ..utils import io_utils


class OmeZarrWriter:
    def __init__(self, uri: types.PathLike):
        """
        Constructor.

        Parameters
        ----------
        uri: types.PathLike
            The URI or local path for where to save the data.
        """
        # Resolve final destination
        fs, path = io_utils.pathlike_to_fs(uri)

        # Save image to zarr store!
        self.store = parse_url(uri, mode="w").store
        self.root_group = zarr.group(store=self.store)

    @staticmethod
    def build_ome(
        size_z: int,
        image_name: str,
        channel_names: List[str],
        channel_colors: List[int],
        channel_minmax: List[Tuple[float, float]],
    ) -> Dict:
        """
        Create the omero metadata for an OME zarr image

        Parameters
        ----------
        size_z:
            Number of z planes
        image_name:
            The name of the image
        channel_names:
            The names for each channel
        channel_colors:
            List of all channel colors
        channel_minmax:
            List of all (min, max) pairs of channel intensities

        Returns
        -------
        Dict
            An "omero" metadata object suitable for writing to ome-zarr
        """
        ch = []
        for i in range(len(channel_names)):
            ch.append(
                {
                    "active": True,
                    "coefficient": 1,
                    "color": f"{channel_colors[i]:06x}",
                    "family": "linear",
                    "inverted": False,
                    "label": channel_names[i],
                    "window": {
                        "end": float(channel_minmax[i][1]),
                        "max": float(channel_minmax[i][1]),
                        "min": float(channel_minmax[i][0]),
                        "start": float(channel_minmax[i][0]),
                    },
                }
            )

        omero = {
            "id": 1,  # ID in OMERO
            "name": image_name,  # Name as shown in the UI
            "version": "0.4",  # Current version
            "channels": ch,
            "rdefs": {
                "defaultT": 0,  # First timepoint to show the user
                "defaultZ": size_z // 2,  # First Z section to show the user
                "model": "color",  # "color" or "greyscale"
            },
            # TODO: can we add more metadata here?
            # # from here down this is all extra and not part of the ome-zarr spec
            # "meta": {
            #     "projectDescription": "20+ lines of gene edited cells etc",
            #     "datasetName": "aics_hipsc_v2020.1",
            #     "projectId": 2,
            #     "imageDescription": "foo bar",
            #     "imageTimestamp": 1277977808.0,
            #     "imageId": 12,
            #     "imageAuthor": "danielt",
            #     "imageName": "AICS-12_143.ome.tif",
            #     "datasetDescription": "variance dataset after QC",
            #     "projectName": "aics cell variance project",
            #     "datasetId": 3
            # },
        }
        return omero

    def write_image(
        self,
        # TODO how to pass in precomputed multiscales?
        image_data: types.ArrayLike,  # must be 3D, 4D or 5D
        image_name: str,
        physical_pixel_sizes: Optional[types.PhysicalPixelSizes],
        channel_names: Optional[List[str]],
        channel_colors: Optional[List[int]],
        scale_num_levels: int = 1,
        scale_factor: float = 2.0,
        dimension_order: Optional[str] = None,
    ) -> None:
        """
        Write a data array to a file.
        NOTE that this API is not yet finalized and will change in the future.

        Parameters
        ----------
        image_data: types.ArrayLike
            The array of data to store. Data arrays must have 2 to 6 dimensions. If a
            list is provided, then it is understood to be multiple images written to the
            ome-tiff file. All following metadata parameters will be expanded to the
            length of this list.
        image_name: str
            string representing the name of the image
        physical_pixel_sizes: Optional[types.PhysicalPixelSizes]
            PhysicalPixelSizes object representing the physical pixel sizes in Z, Y, X
            in microns.
            Default: None
        channel_names: Optional[List[str]]
            Lists of strings representing the names of the data channels
            Default: None
            If None is given, the list will be generated as a 0-indexed list of strings
            of the form "Channel:image_index:channel_index"
        channel_colors: Optional[List[int]]
            List of rgb color values per channel or a list of lists for each image.
            These must be values compatible with the OME spec.
            Default: None
        scale_num_levels: Optional[int]
            Number of pyramid levels to use for the image.
            Default: 1 (represents no downsampled levels)
        scale_factor: Optional[float]
            The scale factor to use for the image. Only active if scale_num_levels > 1.
            Default: 2.0
        dimension_order: Optional[str]
            The dimension order of the data. If None is given, the dimension order will
            be guessed from the number of dimensions in the data according to TCZYX
            order.

        Examples
        --------
        Write a TCZYX data set to OME-Zarr

        >>> image = numpy.ndarray([1, 10, 3, 1024, 2048])
        ... writer = OmeZarrWriter("/path/to/file.ome.zarr")
        ... writer.write_image(image)

        Write multi-scene data to OME-Zarr, specifying channel names

        >>> image0 = numpy.ndarray([3, 10, 1024, 2048])
        ... image1 = numpy.ndarray([3, 10, 512, 512])
        ... writer = OmeZarrWriter("/path/to/file.ome.zarr")
        ... writer.write_image(image0, "Image:0", ["C00","C01","C02"])
        ... writer.write_image(image1, "Image:1", ["C10","C11","C12"])
        """
        ndims = len(image_data.shape)
        if ndims < 2 or ndims > 5:
            raise exceptions.InvalidDimensionOrderingError(
                f"Image data must have 2, 3, 4, or 5 dimensions. "
                f"Received image data with shape: {image_data.shape}"
            )
        if dimension_order is None:
            dimension_order = "TCZYX"[-ndims:]
        if len(dimension_order) != ndims:
            raise exceptions.InvalidDimensionOrderingError(
                f"Dimension order {dimension_order} does not match data "
                f"shape: {image_data.shape}"
            )
        if (len(set(dimension_order) - set("TCZYX")) > 0) or len(
            dimension_order
        ) != len(set(dimension_order)):
            raise exceptions.InvalidDimensionOrderingError(
                f"Dimension order {dimension_order} is invalid or "
                "contains unexpected dimensions. Only TCZYX currently supported."
            )
        xdimindex = dimension_order.find("X")
        ydimindex = dimension_order.find("Y")
        zdimindex = dimension_order.find("Z")
        cdimindex = dimension_order.find("C")
        if cdimindex > min(i for i in [xdimindex, ydimindex, zdimindex] if i > -1):
            raise exceptions.InvalidDimensionOrderingError(
                f"Dimension order {dimension_order} is invalid. Channel dimension "
                "must be before X, Y, and Z."
            )

        if physical_pixel_sizes is None:
            pixelsizes = (1.0, 1.0, 1.0)
        else:
            pixelsizes = (
                physical_pixel_sizes.Z if physical_pixel_sizes.Z is not None else 1.0,
                physical_pixel_sizes.Y if physical_pixel_sizes.Y is not None else 1.0,
                physical_pixel_sizes.X if physical_pixel_sizes.X is not None else 1.0,
            )
        if channel_names is None:
            # TODO this isn't generating a very pretty looking name but it will be
            # unique
            channel_names = (
                [
                    utils.generate_ome_channel_id(image_id=image_name, channel_id=i)
                    for i in range(image_data.shape[cdimindex])
                ]
                if cdimindex > -1
                else [utils.generate_ome_channel_id(image_id=image_name, channel_id=0)]
            )
        if channel_colors is None:
            # TODO generate proper colors or confirm that the underlying lib can handle
            # None
            channel_colors = (
                [i for i in range(image_data.shape[cdimindex])]
                if cdimindex > -1
                else [0]
            )
        scale_dim_map = {
            "T": 1.0,
            "C": 1.0,
            "Z": pixelsizes[0],
            "Y": pixelsizes[1],
            "X": pixelsizes[2],
        }
        transforms = [
            [
                # the voxel size for the first scale level
                {
                    "type": "scale",
                    "scale": [scale_dim_map[d] for d in dimension_order],
                }
            ]
        ]
        # TODO precompute sizes for downsampled also.
        plane_size = (
            image_data.shape[xdimindex]
            * image_data.shape[ydimindex]
            * image_data.itemsize
        )
        target_chunk_size = 16 * (1024 * 1024)  # 16 MB
        # this is making an assumption of chunking whole XY planes.
        # TODO allow callers to configure chunk dims?
        nplanes_per_chunk = int(math.ceil(target_chunk_size / plane_size))
        nplanes_per_chunk = (
            min(nplanes_per_chunk, image_data.shape[zdimindex]) if zdimindex > -1 else 1
        )
        chunk_dim_map = {
            "T": 1,
            "C": 1,
            "Z": nplanes_per_chunk,
            "Y": image_data.shape[ydimindex],
            "X": image_data.shape[xdimindex],
        }
        chunk_dims = [
            dict(
                chunks=tuple(chunk_dim_map[d] for d in dimension_order),
                compressor=default_compressor,
            )
        ]
        lasty = image_data.shape[ydimindex]
        lastx = image_data.shape[xdimindex]
        # TODO scaler might want to use different method for segmentations than raw
        # TODO allow custom scaler or pre-computed multiresolution levels
        if scale_num_levels > 1:
            # TODO As of this writing, this Scaler is not the most general
            # implementation (it does things by xy plane) but it's code already
            # written that also works with dask, so it's a good starting point.
            scaler = Scaler()
            scaler.method = "nearest"
            scaler.max_layer = scale_num_levels - 1
            scaler.downscale = scale_factor if scale_factor is not None else 2
            for i in range(scale_num_levels - 1):
                scale_dim_map["Y"] *= scaler.downscale
                scale_dim_map["X"] *= scaler.downscale
                transforms.append(
                    [
                        {
                            "type": "scale",
                            "scale": [scale_dim_map[d] for d in dimension_order],
                        }
                    ]
                )
                lasty = int(math.ceil(lasty / scaler.downscale))
                lastx = int(math.ceil(lastx / scaler.downscale))
                plane_size = lasty * lastx * image_data.itemsize
                nplanes_per_chunk = int(math.ceil(target_chunk_size / plane_size))
                nplanes_per_chunk = (
                    min(nplanes_per_chunk, image_data.shape[zdimindex])
                    if zdimindex > -1
                    else 1
                )
                chunk_dim_map["Z"] = nplanes_per_chunk
                chunk_dim_map["Y"] = lasty
                chunk_dim_map["X"] = lastx
                chunk_dims.append(
                    dict(
                        chunks=tuple(chunk_dim_map[d] for d in dimension_order),
                        compressor=default_compressor,
                    )
                )
        else:
            scaler = None

        # try to construct per-image metadata
        ome_json = OmeZarrWriter.build_ome(
            image_data.shape[zdimindex] if zdimindex > -1 else 1,
            image_name,
            channel_names=channel_names,  # type: ignore
            channel_colors=channel_colors,  # type: ignore
            # This can be slow if computed here.
            # TODO: Rely on user to supply the per-channel min/max.
            channel_minmax=[
                (0.0, 1.0)
                for i in range(image_data.shape[cdimindex] if cdimindex > -1 else 1)
            ],
        )
        # TODO user supplies units?
        dim_to_axis = {
            "T": {"name": "t", "type": "time", "unit": "millisecond"},
            "C": {"name": "c", "type": "channel"},
            "Z": {"name": "z", "type": "space", "unit": "micrometer"},
            "Y": {"name": "y", "type": "space", "unit": "micrometer"},
            "X": {"name": "x", "type": "space", "unit": "micrometer"},
        }

        axes = [dim_to_axis[d] for d in dimension_order]

        # TODO image name must be unique within this root group
        group = self.root_group  # .create_group(image_name, overwrite=True)
        group.attrs["omero"] = ome_json

        write_image(
            image=image_data,
            group=group,
            scaler=scaler,
            axes=axes,
            # For each resolution, we have a List of transformation Dicts (not
            # validated). Each list of dicts are added to each datasets in order.
            coordinate_transformations=transforms,
            # Options to be passed on to the storage backend. A list would need to
            # match the number of datasets in a multiresolution pyramid. One can
            # provide different chunk size for each level of a pyramid using this
            # option.
            storage_options=chunk_dims,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import xml.etree.ElementTree as ET
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.error import URLError

import xarray as xr
from fsspec.implementations.local import LocalFileSystem
from fsspec.spec import AbstractFileSystem
from ome_types import OME, from_xml
from pydantic.error_wrappers import ValidationError
from tifffile.tifffile import TiffFile, TiffFileError, TiffTags
from xmlschema import XMLSchemaValidationError
from xmlschema.exceptions import XMLSchemaValueError

from .. import constants, exceptions, transforms, types
from ..dimensions import (
    DEFAULT_CHUNK_DIMS,
    DEFAULT_DIMENSION_ORDER,
    DEFAULT_DIMENSION_ORDER_WITH_SAMPLES,
    DimensionNames,
)
from ..metadata import utils as metadata_utils
from ..types import PhysicalPixelSizes
from ..utils import io_utils
from .tiff_reader import TiffReader

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


class OmeTiffReader(TiffReader):
    """
    Wraps the tifffile and ome-types APIs to provide the same aicsimageio Reader
    API but for volumetric OME-TIFF images.
    Parameters
    ----------
    image: types.PathLike
        Path to image file to construct Reader for.
    chunk_dims: List[str]
        Which dimensions to create chunks for.
        Default: DEFAULT_CHUNK_DIMS
        Note: Dimensions.SpatialY, Dimensions.SpatialX, and DimensionNames.Samples,
        will always be added to the list if not present during dask array
        construction.
    clean_metadata: bool
        Should the OME XML metadata found in the file be cleaned for known
        AICSImageIO 3.x and earlier created errors.
        Default: True (Clean the metadata for known errors)
    fs_kwargs: Dict[str, Any]
        Any specific keyword arguments to pass down to the fsspec created filesystem.
        Default: {}
    Notes
    -----
    If the OME metadata in your file isn't OME schema compilant or does not validate
    this will fail to read your file and raise an exception.
    If the OME metadata in your file doesn't use the latest OME schema (2016-06),
    this reader will make a request to the referenced remote OME schema to validate.
    """

    @staticmethod
    def _get_ome(ome_xml: str, clean_metadata: bool = True) -> OME:
        # To clean or not to clean, that is the question
        if clean_metadata:
            ome_xml = metadata_utils.clean_ome_xml_for_known_issues(ome_xml)

        return from_xml(ome_xml, parser="lxml")

    @staticmethod
    def _is_supported_image(
        fs: AbstractFileSystem, path: str, clean_metadata: bool = True, **kwargs: Any
    ) -> bool:
        try:
            with fs.open(path) as open_resource:
                with TiffFile(open_resource) as tiff:
                    # Get first page description (aka the description tag in general)
                    # after Tifffile version 2023.3.15 mmstack images read all scenes
                    # into tiff.pages[0]
                    xml = tiff.pages[0].description
                    ome = OmeTiffReader._get_ome(xml, clean_metadata)

                    # Handle no images in metadata
                    # this commonly means it is a "BinaryData" OME file
                    # i.e. a non-main OME-TIFF from MicroManager or similar
                    # in this case, because it's not the main file we want to just role
                    # back to TiffReader
                    if ome.binary_only:
                        return False

                    return True

        # tifffile exceptions
        except (TiffFileError, TypeError):
            return False

        # xml parse errors
        except ET.ParseError as e:
            log.debug(f"Failed to parse XML for the provided file. Error: {e}")
            return False

        # invalid OME XMl
        except (XMLSchemaValueError, XMLSchemaValidationError, ValidationError) as e:
            log.debug(f"OME XML validation failed. Error: {e}")
            return False

        # cant connect to external schema resource (no internet conection)
        except URLError as e:
            log.debug(
                f"Could not validate OME XML against referenced schema "
                f"(no internet connection). "
                f"Error: {e}"
            )
            return False

        except Exception as e:
            log.debug(f"Unhandled exception: {e}")
            return False

    @staticmethod
    def _guess_ome_dim_order(tiff: TiffFile, ome: OME, scene_index: int) -> List[str]:
        """
        Guess the dimension order based on OME metadata and actual TIFF data.
        Parameters
        -------
        tiff: TiffFile
            A constructed TIFF object to retrieve data from.
        ome: OME
            A constructed OME object to retrieve data from.
        scene_index: int
            The current operating scene index to pull metadata from.
        Returns
        -------
        dims: List[str]
            Educated guess of the dimension order for the file
        """
        dims_from_ome = metadata_utils.get_dims_from_ome(ome, scene_index)

        # Assumes the dimensions coming from here are align semantically
        # with the dimensions specified in this package. Possible T dimension
        # is not equivalent to T dimension here. However, any dimensions
        # not also found in OME will be omitted.
        dims_from_tiff_axes = list(tiff.series[scene_index].axes)

        # Adjust the guess of what the dimensions are based on the combined
        # information from the tiff axes and the OME metadata.
        # Necessary since while OME metadata should be source of truth, it
        # does not provide enough data to guess which dimension is Samples
        # for RGB files
        dims = [dim for dim in dims_from_ome if dim not in dims_from_tiff_axes]
        dims += [dim for dim in dims_from_tiff_axes if dim in dims_from_ome]
        return dims

    def __init__(
        self,
        image: types.PathLike,
        chunk_dims: Union[str, List[str]] = DEFAULT_CHUNK_DIMS,
        clean_metadata: bool = True,
        fs_kwargs: Dict[str, Any] = {},
        **kwargs: Any,
    ):
        # Expand details of provided image
        self._fs, self._path = io_utils.pathlike_to_fs(
            image,
            enforce_exists=True,
            fs_kwargs=fs_kwargs,
        )

        # Store params
        if isinstance(chunk_dims, str):
            chunk_dims = list(chunk_dims)

        self.chunk_dims = chunk_dims
        self.clean_metadata = clean_metadata

        # Enforce valid image
        if not self._is_supported_image(self._fs, self._path, clean_metadata):
            raise exceptions.UnsupportedFileFormatError(
                self.__class__.__name__, self._path
            )

        # Get ome-types object and warn of other behaviors
        with self._fs.open(self._path) as open_resource:
            with TiffFile(open_resource, is_mmstack=False) as tiff:
                # Get and store OME
                self._ome = self._get_ome(
                    tiff.pages[0].description, self.clean_metadata
                )

                # Get and store scenes
                self._scenes: Tuple[str, ...] = tuple(
                    image_meta.id for image_meta in self._ome.images
                )

                # Log a warning stating that if this is a MM OME-TIFF, don't read
                # many series
                if tiff.is_micromanager and not isinstance(self._fs, LocalFileSystem):
                    log.warning(
                        "**Remote reading** (S3, GCS, HTTPS, etc.) of multi-image "
                        "(or scene) OME-TIFFs created by MicroManager has limited "
                        "support with the scene API. "
                        "It is recommended to use independent AICSImage or Reader "
                        "objects for each remote file instead of the `set_scene` API. "
                        "Track progress on support here: "
                        "https://github.com/AllenCellModeling/aicsimageio/issues/196"
                    )

    @property
    def scenes(self) -> Tuple[str, ...]:
        return self._scenes

    @staticmethod
    def _expand_dims_to_match_ome(
        image_data: types.ArrayLike,
        ome: OME,
        dims: List[str],
        scene_index: int,
    ) -> types.ArrayLike:
        # Expand image_data for empty dimensions
        ome_shape = []

        # need to correct channel count if this is a RGB image
        n_samples = ome.images[scene_index].pixels.channels[0].samples_per_pixel
        has_multiple_samples = n_samples is not None and n_samples > 1
        for d in dims:
            # SizeC can represent RGB (Samples) data rather
            # than channel data, whether or not this is the case depends
            # on what the SamplesPerPixel are for the channel
            if d == "C" and has_multiple_samples:
                count = len(ome.images[scene_index].pixels.channels)
            elif d == "S" and has_multiple_samples:
                count = n_samples
            else:
                count = getattr(ome.images[scene_index].pixels, f"size_{d.lower()}")
            ome_shape.append(count)

        # The file may not have all the data but OME requires certain dimensions
        # expand to fill
        expand_dim_ops: List[Optional[slice]] = []
        for d_size in ome_shape:
            # Add empty dimension where OME requires dimension but no data exists
            if d_size == 1:
                expand_dim_ops.append(None)
            # Add full slice where data exists
            else:
                expand_dim_ops.append(slice(None, None, None))

        # Apply operators to dask array
        return image_data[tuple(expand_dim_ops)]

    def _general_data_array_constructor(
        self,
        image_data: types.ArrayLike,
        dims: List[str],
        coords: Dict[str, Union[List[Any], types.ArrayLike]],
        tiff_tags: TiffTags,
    ) -> xr.DataArray:
        # Expand the image data to match the OME empty dimensions
        image_data = self._expand_dims_to_match_ome(
            image_data=image_data,
            ome=self._ome,
            dims=dims,
            scene_index=self.current_scene_index,
        )

        # Always order array
        if DimensionNames.Samples in dims:
            out_order = DEFAULT_DIMENSION_ORDER_WITH_SAMPLES
        else:
            out_order = DEFAULT_DIMENSION_ORDER

        # Transform into order
        image_data = transforms.reshape_data(
            image_data,
            "".join(dims),
            out_order,
        )

        # Reset dims after transform
        dims = [d for d in out_order]

        return xr.DataArray(
            image_data,
            dims=dims,
            coords=coords,
            attrs={
                constants.METADATA_UNPROCESSED: tiff_tags,
                constants.METADATA_PROCESSED: self._ome,
            },
        )

    def _read_delayed(self) -> xr.DataArray:
        """
        Construct the delayed xarray DataArray object for the image.
        Returns
        -------
        image: xr.DataArray
            The fully constructed and fully delayed image as a DataArray object.
            Metadata is attached in some cases as coords, dims, and attrs contains
            unprocessed tags and processed OME object.
        Raises
        ------
        exceptions.UnsupportedFileFormatError
            The file could not be read or is not supported.
        """
        with self._fs.open(self._path) as open_resource:
            with TiffFile(open_resource, is_mmstack=False) as tiff:
                # Get unprocessed metadata from tags
                tiff_tags = self._get_tiff_tags(tiff)

                # Unpack coords from OME
                coords = metadata_utils.get_coords_from_ome(
                    ome=self._ome,
                    scene_index=self.current_scene_index,
                )

                # Guess the dim order based on metadata and actual tiff data
                dims = OmeTiffReader._guess_ome_dim_order(
                    tiff, self._ome, self.current_scene_index
                )

                # Grab the tifffile axes to use for dask array construction
                # If any of the non-"standard" dims are present
                # they will be filtered out during later reshape data calls
                strictly_read_dims = list(tiff.series[self.current_scene_index].axes)

                # Create the delayed dask array
                image_data = self._create_dask_array(tiff, strictly_read_dims)

                return self._general_data_array_constructor(
                    image_data,
                    dims,
                    coords,
                    tiff_tags,
                )

    def _read_immediate(self) -> xr.DataArray:
        """
        Construct the in-memory xarray DataArray object for the image.
        Returns
        -------
        image: xr.DataArray
            The fully constructed and fully read into memory image as a DataArray
            object. Metadata is attached in some cases as coords, dims, and attrs
            contains unprocessed tags and processed OME object.
        Raises
        ------
        exceptions.UnsupportedFileFormatError
            The file could not be read or is not supported.
        """
        with self._fs.open(self._path) as open_resource:
            with TiffFile(open_resource, is_mmstack=False) as tiff:
                # Get unprocessed metadata from tags
                tiff_tags = self._get_tiff_tags(tiff)

                # Unpack coords from OME
                coords = metadata_utils.get_coords_from_ome(
                    ome=self._ome,
                    scene_index=self.current_scene_index,
                )

                # Guess the dim order based on metadata and actual tiff data
                dims = OmeTiffReader._guess_ome_dim_order(
                    tiff, self._ome, self.current_scene_index
                )

                # Read image into memory
                image_data = tiff.series[self.current_scene_index].asarray()

                return self._general_data_array_constructor(
                    image_data,
                    dims,
                    coords,
                    tiff_tags,
                )

    @property
    def ome_metadata(self) -> OME:
        return self.metadata

    @property
    def physical_pixel_sizes(self) -> PhysicalPixelSizes:
        """
        Returns
        -------
        sizes: PhysicalPixelSizes
            Using available metadata, the floats representing physical pixel sizes for
            dimensions Z, Y, and X.
        Notes
        -----
        We currently do not handle unit attachment to these values. Please see the file
        metadata for unit information.
        """
        return metadata_utils.physical_pixel_sizes(
            self.metadata, self.current_scene_index
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py
from pathlib import Path
from typing import Dict, List


class BuildPyCommand(build_py):
    """Check for existence of XSLT before building."""

    def run(self):
        xslt = (
            Path(__file__).parent
            / "aicsimageio/metadata/czi-to-ome-xslt/xslt/czi-to-ome.xsl"
        )
        if not xslt.is_file():
            raise FileNotFoundError("XSLT not found. Is the submodule checked out?")
        build_py.run(self)


with open("README.md") as readme_file:
    readme = readme_file.read()


# If you need to update a version pin for one of these supporting format libs,
# be sure to also check if you need to update these versions in the
# "READER_TO_INSTALL" lookup table from aicsimageio/formats.py.
format_libs: Dict[str, List[str]] = {
    "base-imageio": [
        "imageio[ffmpeg]>=2.11.0,<2.28.0",
        "Pillow>=9.3.0",
    ],
    "nd2": ["nd2[legacy]>=0.2.0"],
    "dv": ["mrc>=0.2.0"],
    "bfio": ["bfio>=2.3.0", "tifffile<2022.4.22"],
    # "czi": [  # excluded for licensing reasons
    #     "fsspec>=2022.8.0",
    #     "aicspylibczi>=3.1.1",
    # ],
    # "bioformats": ["bioformats_jar"],  # excluded for licensing reasons
    # "lif": ["readlif>=0.6.4"],  # excluded for licensing reasons
}

all_formats: List[str] = []
for deps in format_libs.values():
    for dep in deps:
        all_formats.append(dep)

setup_requirements = [
    "pytest-runner>=5.2",
]

test_requirements = [
    "dask[array,distributed]>=2021.4.1,!=2022.5.1",
    "docutils>=0.10,<0.16",
    "psutil>=5.7.0",
    "pytest>=5.4.3",
    "pytest-cov>=2.9.0",
    "pytest-raises>=0.11",
    "quilt3",  # no pin to avoid pip cycling (boto is really hard to manage)
    "s3fs[boto3]>=2022.11.0",
    "tox==3.27.1",
]

dev_requirements = [
    *setup_requirements,
    *test_requirements,
    "asv>=0.4.2",
    "black>=22.3.0",
    "bump2version>=1.0.1",
    "coverage>=5.1",
    "flake8>=3.8.3",
    "flake8-debugger>=3.2.1",
    "gitchangelog>=3.0.4",
    "ipython>=7.15.0",
    "isort>=5.11.5",
    "m2r2>=0.2.7",
    "mypy>=0.800",
    "pytest-runner>=5.2",
    "Sphinx>=3.4.3",
    "sphinx_rtd_theme>=0.5.1",
    "twine>=3.1.1",
    "wheel>=0.34.2",
    # reader deps
    *all_formats,
    "bioformats_jar",  # to test bioformats
    "bfio>=2.3.0",
    "readlif>=0.6.4",  # to test lif
    "aicspylibczi>=3.1.1",  # to test czi
]

benchmark_requirements = [
    *dev_requirements,
    "dask-image>=0.6.0",
]

requirements = [
    "dask[array]>=2021.4.1",
    "fsspec>=2022.8.0",
    "imagecodecs>=2020.5.30",
    "lxml>=4.6,<5",
    "numpy>=1.16,<2",
    "ome-types>=0.3.3",
    "ome-zarr>=0.6.1",
    "wrapt>=1.12",
    "resource-backed-dask-array>=0.1.0",
    "tifffile>=2021.8.30",
    "xarray>=0.16.1,<2023.02.0",
    "xmlschema",  # no pin because it's pulled in from OME types
    "zarr>=2.6,<3",
]

extra_requirements = {
    "setup": setup_requirements,
    "test": test_requirements,
    "dev": dev_requirements,
    "benchmark": benchmark_requirements,
    **format_libs,
    "all": all_formats,
}

setup(
    author="Eva Maxfield Brown, Allen Institute for Cell Science",
    author_email="evamaxfieldbrown@gmail.com, jamie.sherman@gmail.com, bowdenm@spu.edu",
    cmdclass={"build_py": BuildPyCommand},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    description=(
        "Image Reading, Metadata Conversion, and Image Writing for Microscopy Images "
        "in Pure Python"
    ),
    entry_points={},
    install_requires=requirements,
    license="BSD-3-Clause",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="imageio, image reading, image writing, metadata, microscopy, allen cell",
    name="aicsimageio",
    packages=find_packages(
        exclude=[
            "tests",
            "*.tests",
            "*.tests.*",
            "benchmarks",
            "*.benchmarks",
            "*.benchmarks.*",
        ]
    ),
    python_requires=">=3.8",
    setup_requires=setup_requirements,
    test_suite="aicsimageio/tests",
    tests_require=test_requirements,
    extras_require=extra_requirements,
    url="https://github.com/AllenCellModeling/aicsimageio",
    # Do not edit this string manually, always use bumpversion
    # Details in CONTRIBUTING.md
    version="4.11.0",
    zip_safe=False,
)
